use serde::{Deserialize, Serialize};
use serde_json;

#[derive(Serialize, Deserialize, Debug)]
pub struct Actor {
    #[serde(rename = "type")]
    pub object_type: ObjectType,
    pub name: String, // The username
    #[serde(rename = "preferredUsername")]
    pub preferred_username: String, // The @ handle
    pub following: String, // Link to list of users which are followed by them
    pub followers: String, // Link to list of users which follow them
    pub inbox: String, // Link to collection where notes will be posted to
    pub outbox: String, // Link to notes by this user
    pub summary: Option<String>, // Bio
    pub icon: Option<Icon>, // Link to pfp
}

#[derive(Serialize, Deserialize, Debug, PartialEq)]
pub struct Icon {
    #[serde(rename = "type")]
    media_type: String,
    #[serde(rename = "mediaType")]
    media_mime_type: String,
    url: String,
}
#[derive(Serialize, Deserialize, Debug, PartialEq)]
pub enum ObjectType {
    Person,
    Like,
    Note,
    Tombstone,
}
impl std::fmt::Display for ObjectType {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match self {
            ObjectType::Person => write!(f, "Person"),
            ObjectType::Like => write!(f, "Like"),
            ObjectType::Note => write!(f, "Note"),
            ObjectType::Tombstone => write!(f, "Tombstone"),
        }
    }
}

pub enum Object {
    Person(Actor),
    Like,
    Note,
    Tombstone,
}

#[derive(Serialize, Deserialize, Debug, PartialEq)]
pub struct Activity {
    #[serde(rename = "type")]
    pub activity_type: ObjectType,
}

pub fn json_to_actor(input: &String) -> Result<Actor, serde_json::Error> {
    let actor_result: Result<Actor, serde_json::Error> = serde_json::from_str(&input);
    match actor_result {
        Ok(successful_actor) => Ok(successful_actor),
        Err(error) => Err(error),
    }
}

pub fn parse_activity(input: &String) -> Result<Object, String> {
    let json_input: Result<Activity, _> = serde_json::from_str(input);
    if json_input.is_ok() {
        match json_input.as_ref().unwrap().activity_type {
            ObjectType::Person => match json_to_actor(input) {
                Ok(successful_actor) => Ok(Object::Person(successful_actor)),
                Err(error) => Err(format!("Error while trying to parse actor {}", error)),
            },
            _ => {
                return Err(format!(
                    "Activity type {} not implemented",
                    json_input.unwrap().activity_type
                ));
            }
        }
    } else {
        Err("Failed to parse json data".to_string())
    }
}

#[cfg(test)]
mod tests {

    use super::*;
    use std::fs::read_to_string;

    #[test]
    fn parse_json_into_actor() {
        let input_string = read_to_string("tests/test-json-person.json").unwrap();
        let actor: Actor = json_to_actor(&input_string).unwrap();

        assert_eq!(actor.object_type, ObjectType::Person);
        assert_eq!(actor.name, "Sebastian Jambor".to_string());
        assert_eq!(actor.preferred_username, "crepels");
        assert_eq!(
            actor.following,
            "https://mastodon.social/users/crepels/following".to_string()
        );
        assert_eq!(
            actor.followers,
            "https://mastodon.social/users/crepels/followers".to_string()
        );
        assert_eq!(
            actor.inbox,
            "https://mastodon.social/users/crepels/inbox".to_string()
        );
        assert_eq!(
            actor.outbox,
            "https://mastodon.social/users/crepels/outbox".to_string()
        );
        assert_eq!(
            actor.summary,
            Some("<p>I created a systemd playground to help people learn systemd.</p>".to_string())
        );
        assert_eq!(actor.icon,Some(Icon{media_type: "Image".to_string(), media_mime_type: "image/png".to_string(), url: "https://files.mastodon.social/accounts/avatars/108/227/485/389/961/502/original/1213d525278ae01d.png".to_string()}))
    }

    #[test]
    fn detect_activity_type() {
        let input_string = read_to_string("tests/test-json-person.json").unwrap();
        let result = parse_activity(&input_string).unwrap();
        assert!(matches!(result, Object::Person(..)));
    }
}
